#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import OrderedDict

from cord.orm import base_orm


class Model(base_orm.BaseORM):
    """
    Model base ORM.

    ORM:

    """

    DB_FIELDS = OrderedDict([])


class ModelInferenceParams(base_orm.BaseORM):
    """
    Model inference parameters for running models trained via the platform.

    ORM:

    local_file_path,
    conf_thresh,
    iou_thresh,
    device
    detection_frame_range (optional)

    """

    DB_FIELDS = OrderedDict([
        ("files", list),
        ("conf_thresh", float),  # Confidence threshold
        ("iou_thresh", float),  # Intersection over union threshold
        ("device", str),
        ("detection_frame_range", list)
    ])



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cord.orm.label_row import LabelRow
from cord.utils.str_constants import *


def construct_answer_dictionaries(label_row):
    """
    Adds answer object and classification answer dictionaries from a label row if they do not exist.
    Integrity checks are conducted upon saving of labels.

    Args:
        label_row: A label row.

    Returns:
        LabelRow: A label row instance with updated answer dictionaries
    """
    label_row = LabelRow(label_row)  # Cast to label row ORM
    data_type = label_row.data_type
    data_units = label_row.data_units

    object_answers = data_units.get(OBJECT_ANSWERS, {})
    classification_answers = data_units.get(CLASSIFICATION_ANSWERS, {})

    for du in data_units:  # Iterate over data units in label row
        data_unit = data_units[du]

        if LABELS in data_unit:
            labels = data_unit.get(LABELS)

            if data_type == IMG_GROUP:  # Go through images
                items = labels.get(OBJECTS) + labels.get(CLASSIFICATIONS)
                add_answers_to_items(items, classification_answers, object_answers)

            elif data_type == VIDEO:
                for frame in labels:  # Go through frames
                    items = labels[frame].get(OBJECTS) + labels[frame].get(CLASSIFICATIONS)
                    add_answers_to_items(items, classification_answers, object_answers)

    label_row.data_units[OBJECT_ANSWERS] = object_answers
    label_row.data_units[CLASSIFICATION_ANSWERS] = classification_answers
    return label_row


# ---------------------------------------------------------
#                   Helper functions
# ---------------------------------------------------------
def add_answers_to_items(items, classification_answers, object_answers):
    """
    If object_hash (uid) or classification_hash (uid) are not in answer dictionaries,
    add key entry with empty classification list.
    """
    for item in items:
        if OBJECT_HASH in item:
            object_hash = item.get(OBJECT_HASH)
            if object_hash not in object_answers:
                object_answers[object_hash] = {
                    OBJECT_HASH: object_hash,
                    CLASSIFICATIONS: [],
                }

        if CLASSIFICATION_HASH in item:
            classification_hash = item.get(CLASSIFICATION_HASH)
            if classification_hash not in classification_answers:
                classification_answers[classification_hash] = {
                    CLASSIFICATION_HASH: classification_hash,
                    CLASSIFICATIONS: [],
                }



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import logging

from abc import ABCMeta
import cord.exceptions

CORD_ENDPOINT = 'https://api.cord.tech/public'

_CORD_PROJECT_ID = 'CORD_PROJECT_ID'
_CORD_API_KEY = 'CORD_API_KEY'

READ_TIMEOUT = 180  # In seconds
WRITE_TIMEOUT = 180  # In seconds
CONNECT_TIMEOUT = 180  # In seconds

log = logging.getLogger(__name__)


class Config(metaclass=ABCMeta):
    """
    Config defining endpoint, project id, API key, and timeouts.
    """
    def __init__(self,
                 project_id=None,
                 api_key=None,
                 endpoint=CORD_ENDPOINT
                 ):

        if project_id is None:
            if _CORD_PROJECT_ID not in os.environ:
                raise cord.exceptions.AuthenticationError(
                    message="Project ID not provided"
                )

            project_id = os.environ[_CORD_PROJECT_ID]

        if api_key is None:
            if _CORD_API_KEY not in os.environ:
                raise cord.exceptions.AuthenticationError(
                    message="API key not provided"
                )

            api_key = os.environ[_CORD_API_KEY]

        self.project_id = project_id
        self.api_key = api_key

        self.read_timeout = READ_TIMEOUT
        self.write_timeout = WRITE_TIMEOUT
        self.connect_timeout = CONNECT_TIMEOUT

        self.endpoint = endpoint
        self.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'ProjectID': '%s' % project_id,
            'Authorization': '%s' % api_key,
        }

        log.info("Initialised Cord client with endpoint: %s and project_id: %s", endpoint, project_id)


class CordConfig(Config):
    def __init__(self, project_id=None, api_key=None):
        super(CordConfig, self).__init__(project_id, api_key)



import json

from cord.http.query_methods import QueryMethods


class Request(object):
    """
    Request object. Takes query parameters and prepares them for execution.
    """
    def __init__(self,
                 query_method,
                 db_object_type,
                 uid,
                 headers,
                 timeout,
                 connect_timeout,
                 payload,
                 ):
        self.http_method = QueryMethods.POST
        self.data = json.dumps({'query_type': db_object_type.__name__.lower(),
                                'query_method': query_method,
                                'values': {
                                    'uid': uid,
                                    'payload': payload,
                                }})
        self.headers = headers
        self.timeout = timeout
        self.connect_timeout = connect_timeout

    def __str__(self):
        return "Request({}, {}, {}, {})".format(
            self.http_method,
            self.data,
            self.headers,
            self.timeout,
            self.connect_timeout,
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


# Label row constants
LABELS = "labels"
OBJECTS = "objects"
CLASSIFICATIONS = "classifications"
OBJECT_HASH = "objectHash"
CLASSIFICATION_HASH = "classificationHash"
OBJECT_ANSWERS = "object_answers"
CLASSIFICATION_ANSWERS = "classification_answers"

# Data types
VIDEO = "video"
IMG_GROUP = "img_group"

# Labeling algorithm names
INTERPOLATION = "interpolation"



#
# Copyright (c) 2020 Cord Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import datetime
import json

from collections import OrderedDict


class BaseORM(dict):
    """ Base ORM for all database objects. """

    DB_FIELDS = OrderedDict()
    NON_UPDATABLE_FIELDS = {}

    def __init__(self, dic):
        """
        Construct client ORM compatible database object from dict object, ensures strict type and attribute name check
        The real k,v is stored in inner dict
        :param dic:
        """
        try:
            if not isinstance(dic, dict):
                raise TypeError("Need dict object")
            value = {}
            for k, v in dic.items():
                if k in self.DB_FIELDS:
                    types = self.DB_FIELDS[k]
                    # Convert all types to tuple
                    if not isinstance(types, tuple):
                        types = types,
                    # None value is allowed for some cases
                    if v is None:
                        value[k] = v
                    # Normal cases where type matches required types
                    elif isinstance(v, types):
                        value[k] = v
                    # Bool value is same as 0,1 in db
                    elif v in (0, 1) and bool in types:
                        value[k] = v
                    # Datetime type but actually a datetime str is provided
                    elif datetime.datetime in types:
                        real_v = datetime.datetime.strptime(v, "%Y-%m-%d %H:%M:%S")
                        value[k] = real_v
                    elif dict in types:
                        value[k] = v
            super().__init__(**value)
        except Exception as e:
            logging.error("Error init", exc_info=True)
            raise Exception("Convert failed {}".format(str(e)))

    def __getattr__(self, name):
        """
        Override attribute method for easy access of field value in business logic instead of ["attr"]
        Return None if there is no such attribute
        :param name:
        :return:
        """
        if name in self:
            try:
                return self[name]
            except KeyError as e:
                return None
        else:
            raise AttributeError("Attribute does not exist: {}".format(name))

    def __setattr__(self, name, value):
        """
        Strict attribute name and type check
        :param name:
        :param value:
        :return:
        """
        if name in self.DB_FIELDS and (value is None or isinstance(value, self.DB_FIELDS[name])):
            self[name] = value
        else:
            raise AttributeError("Attribute name or type not match: {}".format(name))

    def __delattr__(self, name):
        if name in self and name in self.DB_FIELDS:
            del self[name]
        else:
            super().__delattr__(self, name)

    @staticmethod
    def from_db_row(row, db_field):
        """
        Static method for conveniently converting db row to client object
        :param row:
        :param db_field:
        :return:
        """
        temp_dict = {}
        for i, attribute in enumerate(db_field):
            temp_dict[attribute] = row[i]
        return temp_dict

    def to_dic(self, time_str=True):
        """
        Conveniently set client object as dict
        Only considers the dict items, no other object attr will be counted
        :param time_str: time_str, if set to True, will convert datetime field to str with format %Y-%m-%d %H:%M:%S
        If False, will keep the original datetime type. By default will be True.
        :return:
        """
        res = {}
        for k, v in self.items():
            if isinstance(v, datetime.datetime) and time_str is True:
                v = v.strftime("%Y-%m-%d %H:%M:%S")
            elif isinstance(v, dict):
                v = json.dumps(v)
            res[k] = v

        return res

    def updatable_fields(self):
        for k, v in self.items():
            if k not in self.NON_UPDATABLE_FIELDS and v is not None:
                yield k, v
